package org.xwiki.cache.jbosscache.internal;

import java.util.Map;

import javax.swing.event.EventListenerList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xwiki.cache.CacheNeedsRefreshException;

import org.jboss.cache.config.Configuration;
import org.jboss.cache.notifications.annotation.NodeModified;
import org.jboss.cache.notifications.event.NodeModifiedEvent;
import org.jboss.cache.Cache;
import org.jboss.cache.Fqn;

public class JBossCacheCache implements org.xwiki.cache.Cache
{
    private static final Log LOG = LogFactory.getLog(JBossCacheCache.class);

    private static final Fqn<String> FQN = Fqn.fromString("/xwiki");

    private Cache<String, Object> cache;

    private final EventListenerList cacheEntryListeners = new EventListenerList();

    private Map<String, Object> preEventData;

    public JBossCacheCache(Cache<String, Object> cache)
    {
        this(null, cache);
    }

    public JBossCacheCache(String name, Cache<String, Object> cache)
    {
        this.cache = cache;
        this.cache.addCacheListener(FQN, this);

        if (name != null) {
            Configuration config = cache.getConfiguration();
            config.setClusterName(name);
        }

        cache.create();
        cache.start();
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#addCacheEntryListener(org.xwiki.cache.CacheEntryListener)
     */
    public void addCacheEntryListener(org.xwiki.cache.CacheEntryListener listener)
    {
        cacheEntryListeners.add(org.xwiki.cache.CacheEntryListener.class, listener);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeCacheEntryListener(org.xwiki.cache.CacheEntryListener)
     */
    public void removeCacheEntryListener(org.xwiki.cache.CacheEntryListener listener)
    {
        cacheEntryListeners.remove(org.xwiki.cache.CacheEntryListener.class, listener);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#setCapacity(int)
     */
    public void setCapacity(int capacity)
    {
        // not support
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#remove(java.lang.String)
     */
    public void remove(String key)
    {
        cache.remove(FQN, key);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#set(java.lang.String, java.lang.Object)
     */
    public void set(String key, Object obj)
    {
        cache.put(FQN, key, obj);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String)
     */
    public Object get(String key) throws CacheNeedsRefreshException
    {
        Object data = cache.get(FQN, key);

        if (data == null) {
            throw new CacheNeedsRefreshException(null, data);
        }

        return data;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String, int)
     */
    public Object get(String key, int refeshPeriod) throws CacheNeedsRefreshException
    {
        return cache.get(FQN, key/* , obj */);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#getSize()
     */
    public int getSize()
    {
        return cache.getNode(FQN).dataSize();
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#cancelUpdate(java.lang.String)
     */
    public void cancelUpdate(String key)
    {
        cache.remove(FQN, key);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeAll()
     */
    public void removeAll()
    {
        cache.getNode(FQN).clearData();
    }

    /**
     * {@inheritDoc}
     * 
     * @see java.lang.Object#finalize()
     */
    public void finalize() throws Throwable
    {
        cache.stop();
        cache.destroy();

        super.finalize();
    }

    // ////////////////////////////////////////////////////////////////
    // Events
    // ////////////////////////////////////////////////////////////////

    @NodeModified
    public void nodeModified(NodeModifiedEvent event)
    {
        if (event.isPre()) {
            preEventData = event.getData();
        } else {
            Map<String, Object> data = event.getData();

            for (Map.Entry<String, Object> entry : data.entrySet()) {
                String key = entry.getKey();
                Object value = entry.getValue();

                if (event.getModificationType() == NodeModifiedEvent.ModificationType.REMOVE_DATA) {
                    cacheEntryRemoved(key, value);
                } else {
                    cacheEntryAdded(key, value);
                }
            }
        }
    }

    private void cacheEntryAdded(String key, Object value)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {

            if (preEventData.containsKey(key)) {
                listener.cacheEntryModified(new JBossCacheCacheEntryEvent(this, key, value));
            } else {
                listener.cacheEntryAdded(new JBossCacheCacheEntryEvent(this, key, value));
            }
        }
    }

    private void cacheEntryRemoved(String key, Object value)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {
            listener.cacheEntryRemoved(new JBossCacheCacheEntryEvent(this, key, value));
        }
    }
}
