package org.xwiki.cache.oscache.internal;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class AbstractOSCacheCacheFactory implements Runnable
{
    /**
     * This component's role hint, used when code needs to look it up.
     */
    public final static String ROLEHINT = "oscache";

    private static final Log LOG = LogFactory.getLog(AbstractOSCacheCacheFactory.class);

    private static final String PROPS_PATH = "/WEB-INF/";

    org.xwiki.container.Container container;

    private static int cacheCount;

    /**
     * Used by watcher thread to keep an eye on caches.
     */
    private List<WeakReference<OSCacheCache>> cacheList =
        new LinkedList<WeakReference<OSCacheCache>>();

    protected Properties mergeProperties(Properties props, Properties sourceProps)
    {
        Properties targetProps = new Properties();

        targetProps.putAll(sourceProps);
        targetProps.putAll(props);

        return targetProps;
    }

    protected void initCache(OSCacheCache cache)
    {
        cacheCount++;
        cache.setName("Cache # " + cacheCount);

        if (LOG.isInfoEnabled()) {
            DateFormat df = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");
            LOG.info("Created " + cache.getName() + " of size " + cache.getCapacity() + " at "
                + df.format(new Date()) + ", current count is " + cacheCount);
        }

        synchronized (cacheList) {
            cacheList.add(new WeakReference<OSCacheCache>(cache));
        }
    }

    protected Properties loadProps(String propertiesFilename)
    {
        LOG.info("Loading cache properties: " + propertiesFilename);

        Properties props = new Properties();
        InputStream is = null;

        try {
            File f = new File(propertiesFilename);
            if (f.exists()) {
                is = new FileInputStream(f);
            } else {
                is =
                    container.getApplicationContext().getResourceAsStream(
                        PROPS_PATH + propertiesFilename);
            }
            props.load(is);

            if (LOG.isInfoEnabled()) {
                LOG.info("Properties loaded: " + propertiesFilename);
            }
        } catch (Exception e) {
            LOG.warn("Could not load cache properties " + propertiesFilename + ": "
                + e.getMessage());
        } finally {
            try {
                is.close();
            } catch (IOException e) {
                LOG.debug("Failed t close properties file", e);
            }
        }

        return props;
    }

    /**
     * The cache watcher thread.
     */
    public void run()
    {
        while (true) {
            if (LOG.isInfoEnabled()) {
                synchronized (cacheList) {
                    for (WeakReference<OSCacheCache> wr : cacheList) {
                        if (wr != null) {
                            OSCacheCache cache = (OSCacheCache) wr.get();
                            if (cache != null) {
                                LOG.info("OSCacheCache item count for " + cache.getName() + " = "
                                    + cache.getSize() + " capacity is "
                                    + cache.getCapacity());
                            }
                        }
                    }
                }
            }
            try {
                // 5 min
                Thread.sleep(1000 * 60 * 5);
            } catch (InterruptedException e) {
                LOG.debug("Cache watcher thread has been interrupted", e);
            }

        }
    }
}
