package org.xwiki.cache.oscache.internal;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.Properties;

import javax.swing.event.EventListenerList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xwiki.cache.FinalizableCacheData;
import org.xwiki.cache.CacheException;
import org.xwiki.cache.CacheNeedsRefreshException;

import com.opensymphony.oscache.base.Cache;
import com.opensymphony.oscache.base.CacheEntry;
import com.opensymphony.oscache.base.NeedsRefreshException;
import com.opensymphony.oscache.base.events.CacheEntryEvent;
import com.opensymphony.oscache.base.events.CacheEntryEventListener;
import com.opensymphony.oscache.base.events.CacheGroupEvent;
import com.opensymphony.oscache.base.events.CachePatternEvent;
import com.opensymphony.oscache.base.events.CachewideEvent;
import com.opensymphony.oscache.general.GeneralCacheAdministrator;

public class OSCacheCache implements org.xwiki.cache.Cache, CacheEntryEventListener
{
    private static final Log LOG = LogFactory.getLog(OSCacheCache.class);

    private GeneralCacheAdministrator cacheAdmin;

    private String name;

    private int capacity;

    private final EventListenerList cacheEntryListeners = new EventListenerList();

    public OSCacheCache(Properties props)
    {
        cacheAdmin = new GeneralCacheAdministrator(props);

        cacheAdmin.getCache().addCacheEventListener(this, CacheEntryEventListener.class);
    }

    public OSCacheCache(Properties props, int capacity) throws CacheException
    {
        this(props);

        this.cacheAdmin.setCacheCapacity(capacity);
        this.capacity = capacity;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#addCacheEntryListener(org.xwiki.cache.CacheEntryListener)
     */
    public void addCacheEntryListener(org.xwiki.cache.CacheEntryListener listener)
    {
        cacheEntryListeners.add(org.xwiki.cache.CacheEntryListener.class, listener);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeCacheEntryListener(org.xwiki.cache.CacheEntryListener)
     */
    public void removeCacheEntryListener(org.xwiki.cache.CacheEntryListener listener)
    {
        cacheEntryListeners.remove(org.xwiki.cache.CacheEntryListener.class, listener);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#remove(java.lang.String)
     */
    public void remove(String key)
    {
        this.cacheAdmin.flushEntry(key);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#set(java.lang.String, java.lang.Object)
     */
    public void set(String key, Object obj)
    {
        this.cacheAdmin.putInCache(key, obj);
        logCacheAdd(key, obj);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String)
     */
    public Object get(String key) throws CacheNeedsRefreshException
    {
        try {
            return this.cacheAdmin.getFromCache(key);
        } catch (NeedsRefreshException e) {
            throw new CacheNeedsRefreshException(e, e.getCacheContent());
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#get(java.lang.String, int)
     */
    public Object get(String key, int refeshPeriod) throws CacheNeedsRefreshException
    {
        try {
            return this.cacheAdmin.getFromCache(key, refeshPeriod);
        } catch (NeedsRefreshException e) {
            throw new CacheNeedsRefreshException(e, e.getCacheContent());
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#cancelUpdate(java.lang.String)
     */
    public void cancelUpdate(String key)
    {
        this.cacheAdmin.cancelUpdate(key);
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.xwiki.cache.Cache#removeAll()
     */
    public void removeAll()
    {
        this.cacheAdmin.flushAll();
        Cache cache = this.cacheAdmin.getCache();

        // Make sure we clean-up and finalize cached objects
        Map< ? , ? > map = (Map< ? , ? >) getPrivateField(cache, "cacheMap");

        for (Object value : map.values()) {
            try {
                CacheEntry centry = (CacheEntry) value;
                Object obj = centry.getContent();
                if (obj instanceof FinalizableCacheData) {
                    ((FinalizableCacheData) obj).finalize();
                }
            } catch (Throwable throwable) {
            }
        }

        callPrivateMethod(cache, "clear");
    }

    /**
     * Returns the value of the field represented by this <code>Field</code>, on the specified
     * object. The value is automatically wrapped in an object if it has a primitive type.
     * 
     * @param obj the object from which the represented field's value is to be extracted
     * @param fieldName the name of the field
     * @return the value of the represented field in object <tt>obj</tt>; primitive values are
     *         wrapped in an appropriate object before being returned
     * @see Class#getDeclaredField(String)
     * @see Field#get(Object)
     */
    private static Object getPrivateField(Object obj, String fieldName)
    {
        Object value;

        try {
            Field field = obj.getClass().getDeclaredField(fieldName);
            field.setAccessible(true);

            value = field.get(obj);
        } catch (NoSuchFieldException e) {
            value = null;
        } catch (IllegalAccessException e) {
            LOG.error("Can't access field [" + fieldName + "] of class [" + obj.getClass() + "]",
                e);
            value = null;
        }

        return value;
    }

    /**
     * Invokes the underlying non public method represented by this <code>Method</code> object, on
     * the specified object.
     * 
     * @param obj the object the underlying method is invoked from
     * @param methodName the name of the method
     * @return the result of dispatching the method represented by this object on <code>obj</code>
     *         with parameters <code>args</code>
     * @see Class#getDeclaredMethod(String, Class...)
     * @see Method#invoke(Object, Object...)
     */
    private static Object callPrivateMethod(Object obj, String methodName)
    {
        Object returnValue;

        try {
            Method method = obj.getClass().getDeclaredMethod(methodName);
            method.setAccessible(true);

            returnValue = method.invoke(obj);
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            returnValue = null;
        } catch (NoSuchMethodException e) {
            returnValue = null;
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            returnValue = null;
        }

        return returnValue;
    }

    /**
     * {@inheritDoc}
     * <p>
     * Method called here is labeled "for test only" in OSCache JavaDoc, so don't call this a lot!
     * </p>
     * 
     * @see org.xwiki.cache.Cache#getSize()
     */
    public int getSize()
    {
        return this.cacheAdmin.getCache().getSize();
    }

    public void setName(String name)
    {
        this.name = name;
    }

    public String getName()
    {
        return name;
    }

    public int getCapacity()
    {
        return this.capacity;
    }

    private void logCacheAdd(String sKey, Object obj)
    {
        if (LOG.isDebugEnabled()) {
            StringBuffer sbEntry = new StringBuffer();
            sbEntry.append(getName());
            sbEntry.append(" adding object key=");
            sbEntry.append(sKey);
            sbEntry.append("; class=");
            if (obj == null) {
                sbEntry.append("null");
            } else {
                sbEntry.append(obj.getClass().toString());
                sbEntry.append("; toString=");
                sbEntry.append(obj.toString());
                LOG.debug(sbEntry.toString());
            }
        }
    }

    // ////////////////////////////////////////////////////////////////
    // Events
    // ////////////////////////////////////////////////////////////////

    public void cacheEntryAdded(CacheEntryEvent event)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {
            listener.cacheEntryAdded(new OSCacheCacheEntryEvent(this, event));
        }
    }

    public void cacheEntryFlushed(CacheEntryEvent event)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {
            listener.cacheEntryFlushed(new OSCacheCacheEntryEvent(this, event));
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryRemoved(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryRemoved(CacheEntryEvent event)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {
            listener.cacheEntryRemoved(new OSCacheCacheEntryEvent(this, event));
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheEntryUpdated(com.opensymphony.oscache.base.events.CacheEntryEvent)
     */
    public void cacheEntryUpdated(CacheEntryEvent event)
    {
        for (org.xwiki.cache.CacheEntryListener listener : cacheEntryListeners
            .getListeners(org.xwiki.cache.CacheEntryListener.class)) {
            listener.cacheEntryModified(new OSCacheCacheEntryEvent(this, event));
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheGroupFlushed(com.opensymphony.oscache.base.events.CacheGroupEvent)
     */
    public void cacheGroupFlushed(CacheGroupEvent event)
    {
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cachePatternFlushed(com.opensymphony.oscache.base.events.CachePatternEvent)
     */
    public void cachePatternFlushed(CachePatternEvent event)
    {
    }

    /**
     * {@inheritDoc}
     * 
     * @see com.opensymphony.oscache.base.events.CacheEntryEventListener#cacheFlushed(com.opensymphony.oscache.base.events.CachewideEvent)
     */
    public void cacheFlushed(CachewideEvent event)
    {
    }
}
